package sockprog.example4_timeout;

import java.io.BufferedReader;
import java.io.BufferedWriter;
import java.io.IOException;
import java.io.InputStream;
import java.io.InputStreamReader;
import java.io.OutputStream;
import java.io.OutputStreamWriter;
import java.io.Writer;
import java.net.Socket;
import java.net.SocketTimeoutException;


public final class MessageClient {
    private static final int NO_TIMEOUT = 0;
    
    /**
     * クライアントプログラムを起動する。
     * 
     * <pre>
     * > java sockprog.example4_timeout.MessageClient -t 10 localhost 8888 "Hello, World!"
     * </pre>
     * 
     * のように-t タイムアウト値（秒単位）、
     * サーバーアドレス、ポート番号、送信する文字列の
     * 順に引数を指定する。
     */
    public static void main(String[] args) {
        int timeout = NO_TIMEOUT;

        int offset = 0;
        for(offset = 0; offset < args.length;) {
            if(args.length >= offset + 2 && args[offset].equals("-t")) {
                timeout = Integer.parseInt(args[offset + 1]) * 1000;
                offset += 2;
                continue;
            }
            if(args.length >= offset + 1 && args[offset].startsWith("-")) {
                System.err.println("unknown option: " + args[offset]);
                System.exit(-1);
            }
            break;
        }
        
        if(args.length < offset+3) {
            System.err.println("parameter not enough");
            System.exit(-1);
        }

        String host = args[offset+0];
        int port = Integer.parseInt(args[offset+1]);
        String message = args[offset+2];
        
        MessageClient client = null;
        try {
            client = new MessageClient(host, port, timeout);
            String reply = client.send(message);
            System.out.println(reply);
        } catch (SocketTimeoutException e) {  // [1]
            // タイムアウトの際の処理を行う
            System.err.println("timeout");
        } catch (IOException e) {
            e.printStackTrace();
        } finally {
            try {
                if(client != null) {
                    client.shutDown();
                }
            } catch (IOException e) {
                e.printStackTrace();
            }
        }
    }
    
    private final Socket socket;
    private final Writer writer;
    private final BufferedReader reader;    

    public MessageClient(String host, int port, int timeout)
    throws IOException {
        socket = new Socket(host, port);

        // タイムアウトの設定
        socket.setSoTimeout(timeout); // [2]

        // writerの準備
        OutputStream out = socket.getOutputStream();
        OutputStreamWriter outWriter = 
            new OutputStreamWriter(out, ProtocolConstants.ENCODING);
        writer = new BufferedWriter(outWriter);

        // readerの準備
        InputStream in = socket.getInputStream();
        InputStreamReader inReader =
            new InputStreamReader(in, ProtocolConstants.ENCODING);
        reader = new BufferedReader(inReader);
    }
    public void shutDown() throws IOException {
        socket.close();
    }
    
    public String send(String message) throws IOException {
        // 文字列の送信
        writer.write(message);
        writer.write(ProtocolConstants.LINE_SEPARATOR);
        writer.flush();

        // 文字列の受信
        return reader.readLine();
    }
    
}
