package sockprog.example5_buffer;

import java.io.IOException;
import java.net.InetSocketAddress;
import java.net.ServerSocket;
import java.net.Socket;

public final class MessageServer {
    private static final int DEFAULT_BUFFER_SIZE = 32 * 1024;
    private static final int NO_TIMEOUT = 0;
    
    /**
     * サーバープログラムを起動する
     * 
     * <pre>
     * > java sockprog.example5_buffer.MessageServer -t 10 -b 128 80
     * </pre>
     * のように-t タイムアウト値（秒単位）、-b バッファサイズ（KB単位）、
     * ポート番号の順に引数を指定する
     */
    public static void main(String[] args) {
        int timeout = NO_TIMEOUT;
        int bufferSize = DEFAULT_BUFFER_SIZE;

        int offset = 0;
        for(offset = 0; offset < args.length;) {
            if(args.length >= offset + 2 && args[offset].equals("-t")) {
                timeout = Integer.parseInt(args[offset + 1]) * 1000;
                offset += 2;
                continue;
            }
            if(args.length >= offset + 2 && args[offset].equals("-b")) {
                bufferSize = Integer.parseInt(args[offset + 1]) * 1024;
                offset += 2;
                continue;
            }
            if(args.length >= offset + 1 && args[offset].startsWith("-")) {
                System.err.println("unknown option: " + args[offset]);
                System.exit(-1);
            }
            break;
        }

        if(args.length < offset+1) {
            System.err.println("parameter not enough");
            System.exit(-1);
        }
        int port = Integer.parseInt(args[offset+0]);
        
        try {
            MessageServer server = new MessageServer(port, timeout, bufferSize);  
            server.start();
        } catch (IOException e) {
            e.printStackTrace();
            System.exit(-1);
        }
    }

    private final ServerSocket serverSocket;
    private final int timeout;
    private final int bufferSize;
    private final Thread thread;

    public MessageServer(int port, int timeout,
            int bufferSize) throws IOException {
        serverSocket = new ServerSocket();
        serverSocket.setReuseAddress(true);

        // 受信バッファサイズの設定
        serverSocket.setReceiveBufferSize(bufferSize); // [1]
        
        serverSocket.bind(new InetSocketAddress(port));
        this.timeout = timeout;
        this.bufferSize = bufferSize;

        thread = new Thread(new Runnable() {
            public void run() {
                service();
            }
        });
    }
    
    public void start() {
        thread.start();
    }
    
    private void service() {
        while(true) {
            try {
                Socket socket = serverSocket.accept();
                new EchoClientHandler(socket, timeout, bufferSize).run();   
            } catch(IOException e) {
                return;
            }
        }
    }
    
    public void shutDown()
    throws IOException, InterruptedException {
        serverSocket.close();
        thread.join();
    }

}
