package sockprog.example6_thread;

import java.io.IOException;
import java.net.InetSocketAddress;
import java.net.ServerSocket;
import java.net.Socket;
import java.util.concurrent.ExecutorService;
import java.util.concurrent.Executors;
import java.util.concurrent.TimeUnit;

public final class MessageServer {
    private static final int NO_TIMEOUT = 0;
    private static final int DEFAULT_BUFFER_SIZE = 32 * 1024;
    private static final int DEFAULT_THREAD_POOL_SIZE = 5;
    private static final int TERMINATION_TIMEOUT = 60;

    /**
     * サーバープログラムを起動する
     * 
     * <pre>
     * java sockprog.example6_thread.MessageServer \
     * [-t タイムアウト] [-b バッファサイズ] [-p スレッドプールサイズ] ポート番号
     * </pre>
     * 
     * のように起動する。 上記で[]で括られたものはオプションである
     */
    public static void main(String[] args) {
        int timeout = NO_TIMEOUT;
        int bufferSize = DEFAULT_BUFFER_SIZE;
        int threadPoolSize = DEFAULT_THREAD_POOL_SIZE;

        int offset = 0;
        for(offset = 0; offset < args.length;) {
            if(args.length >= offset + 2 && args[offset].equals("-t")) {
                timeout = Integer.parseInt(args[offset + 1]) * 1000;
                offset += 2;
                continue;
            }
            if(args.length >= offset + 2 && args[offset].equals("-b")) {
                bufferSize = Integer.parseInt(args[offset + 1]) * 1024;
                offset += 2;
                continue;
            }
            if(args.length >= offset + 2 && args[offset].equals("-p")) {
                threadPoolSize = Integer.parseInt(args[offset + 1]);
                offset += 2;
                continue;
            }
            if(args.length >= offset + 1 && args[offset].startsWith("-")) {
                System.err.println("unknown option: " + args[offset]);
                System.exit(-1);
            }
            break;
        }

        if(args.length < offset + 1) {
            System.err.println("parameter not enough");
            System.exit(-1);
        }
        int port = Integer.parseInt(args[offset]);

        try {
            MessageServer server = new MessageServer(port, timeout, bufferSize,
                    threadPoolSize);
            server.start();
        } catch (IOException e) {
            e.printStackTrace();
            System.exit(-1);
        }
    }

    private final ServerSocket serverSocket;
    private final int timeout;
    private final int bufferSize;
    private final Thread thread;
    private final ExecutorService executor; // [1]

    public MessageServer(int port, int timeout,
            int bufferSize, int threadPoolSize) 
    throws IOException {
        serverSocket = new ServerSocket();
        serverSocket.setReuseAddress(true);

        // 受信バッファサイズの設定
        serverSocket.setReceiveBufferSize(bufferSize);

        serverSocket.bind(new InetSocketAddress(port));
        this.timeout = timeout;
        this.bufferSize = bufferSize;

        // スレッドプールの生成
        this.executor = 
            Executors.newFixedThreadPool(threadPoolSize); // [2]

        thread = new Thread(new Runnable() {
            public void run() {
                service();
            }
        });
    }

    public void start() {
        thread.start();
    }

    private void service() {
        while(true) {
            try {
                Socket socket = serverSocket.accept();
                Runnable handler =
                    new EchoClientHandler(socket, timeout, bufferSize);
                executor.execute(handler); // [3]
            } catch (IOException e) {
                return;
            }
        }
    }

    public boolean shutDown()
    throws IOException, InterruptedException {
        serverSocket.close();
        thread.join();
        executor.shutdown(); // [4]
        return executor.awaitTermination(TERMINATION_TIMEOUT,
                TimeUnit.SECONDS); // [5]
    }
}
