#!/usr/bin/perl
# -*- coding: utf-8 -*-
# ↑Emacsで文字コードを強制するときのタグです。

### 受信用スクリプトです。なお、Cygwinではcygwin1.dllの関係でIPv6が使えません。

# 警告レベルを上げるpragmaです。Perlでコーディングするときは必須です。
use strict;
use warnings;

# 今回主役のモジュールです。
use Socket6;
use IO::Socket::Multicast6;

# 1秒未満でsleepしたいときは以下を使いましょう。selectのタイムアウトによる
# スリープは、後方互換性の維持以外に使う理由はありません。
use Time::HiRes qw(usleep);

# 単に「YYYY-MM-DD hh:mm:ss」のような文字列が欲しいだけなら、以下のstrftime()
# を使いましょう。localtime()の返り値を一所懸命処理する必要はありません。
# POSIXと書いてますが、Windowsでも使えます。
use POSIX qw(strftime);
# とは言え、Perlのドキュメントだけだとフォーマットがさっぱりな罠が。そんな
# ときには、Cの関数strftime()を調べましょう。「strftime」でぐぐれば山ほど
# 出てきます。

# スレッド用のモジュールです。
# 本文にも書きましたが、
#   http://perldoc.jp/
# の和訳ドキュメントを大いに参考にさせていただきました。
use threads;
use threads::shared;

# 今回定義した定数の読み込みです。
use MCConfig;

# Perlでは、スレッド間で共有するデータを明示的に指定する必要があります。
my %msgs : shared;

# 表示処理です。
sub display {
    while (1) {
	my $now = time;
	my $expire_time = $now - (DISP_EXPIRES / 1000.0);
	print strftime("Perl [%Y-%m-%d %H:%M:%S]\n", localtime($now));
	{
	    # 排他ロックしてデータを読み書きします。
	    lock(%msgs);
	    foreach my $msg (sort {$a->{name} cmp $b->{name}} values %msgs) {
		if ($msg->{reg_time} < $expire_time) {
		    delete($msgs{$msg->{name}});
		} else {
		    printf("  %04d %s (%s/%d)\n",
			   @{$msg}{qw(cnt name addr port)});
		}
	    }
	}
	print "\n";
	usleep(DISP_INTERVAL * 1000); # ミリ秒→マイクロ秒
    }
}

# 標準出力をノンバッファリングモードに指定します。
$| = 1;

# ・引数でIPv4/IPv6を切り替えます。要素の数が0のとき、いきなり$ARGV[0]
#   と書くと、use warningsの効果で怒られます。
my ($domain, $addr);
if (@ARGV && $ARGV[0] eq '6') {
    print "Mode: IPv6\n\n";
    $domain = AF_INET6;
    $addr = MULTICAST_ADDR6;
} else {
    print "Mode: IPv4\n\n";
    $domain = AF_INET;
    $addr = MULTICAST_ADDR4;
}

# 表示処理を別スレッドで起動します。
threads->create('display');

# 受信処理は現在のスレッドで行います。
my $msock = IO::Socket::Multicast6->new(
    Proto => 'udp',
    Domain => $domain,
    LocalPort => MULTICAST_PORT,
    ReuseAddr => 1)
  or die "Couldn't initialize multicast socket: $!";
$msock->mcast_add($addr) or die "Couldn't set group: $!";
while (1) {
    my $saddr = $msock->recv(my $data, RECV_BUF_SIZE);
    # 共有データを構築します。
    # 詳細は「perldoc threads::shared」を参照してください。
    my $msg = &share({}); # 何故か'&'は省略できないようです。
    @{$msg}{qw(cnt name)} = split(/\s+/, $data, 2);
    @{$msg}{qw(addr port)} = getnameinfo($saddr,
					 NI_NUMERICHOST|NI_NUMERICSERV);
    $msg->{reg_time} = time;
    # 排他ロックしてデータを書き込みます。
    { lock(%msgs); $msgs{$msg->{name}} = $msg }
}
