%%% -*- coding: utf-8 -*-
%%% ↑Emacsで文字コードを強制するときのタグです。

%%% モジュール名はファイルのベース名と一致させます。
%%% モジュール名には、文字列ではなくアトム(≒シンボル)を指定します。
%%% 小文字から始まる名前がアトムになります。
-module(display).

%%% モジュール名を付けずに関数を呼び出すにはimportします。
%%% モジュール名を明示的に指定する場合はimportする必要はありません。
%%% 関数名の後ろの数字は引数の数(arityと言う)です。
-import(lists, [sort/1, foreach/2]).
-import(calendar, [now_to_datetime/1]).
-import(timer, [now_diff/2]).

%%% 外部から呼び出し可能な関数を指定します。
-export([start/0, stop/0, send/4]).

%%% includeはマクロプロセッサで処理されます。escriptでは使えないようです。
-include("mcconfig.hrl").

%%% 表示処理です。
disp(Name, {Cnt, Addr, Port, RegTime}, Now) ->
    After = now_diff(Now, RegTime), % 大文字から始まる名前は変数です。
    %% ?NAMEはマクロ展開です。
    if After > ?DISP_EXPIRES * 1000 ->
	    erase(Name);
       %% 小文字から始まる名前はアトム(≒シンボル)です。
       %% 実は、trueやfalseという予約語があるわけではありません。
       true ->
	    %% io:formatはprintf相当です。フォーマットについてはioモジュールの
	    %% ドキュメントを参照してください。'['…']'はリストです。
	    io:format("  ~4.10.0B ~s (~p/~p)~n", [Cnt, Name, Addr, Port])
    end.

%%% 表示用プロセスのメインループです。
loop(Prev) ->
    Start = now(),
    %% マイクロ秒→ミリ秒。divは整数除算です。
    Diff = ?DISP_INTERVAL - (now_diff(Start, Prev) div 1000),
    Timeout = if Diff =< 0 -> 0; true -> Diff end,
    receive
        {Cnt, Name, Addr, Port} -> % 受信プロセスからのメッセージを受信します。
            put(Name, {Cnt, Addr, Port, now()}), % '{'…'}'はタプルです。
            loop(Prev);
        stop -> % stopシンボルを受信したらプロセスを停止します。
            void
    after Timeout -> % タイムアウトしたら受信状況を出力します。
            Now = now(),
            {{Y, Mo, D}, {H, M, S}} = now_to_datetime(Now),
	    %% リストや文字列の結合には++を用います。
	    %% 文字列の実体は文字のリストです。
            io:format("Erlang [~4.10.0B-~2.10.0B-~2.10.0B" ++
                      " ~2.10.0B:~2.10.0B:~2.10.0B]~n",
                      [Y, Mo, D, H, M, S]),
            foreach(fun({Name, Entry}) ->
			    disp(Name, Entry, Now)
		    end, sort(get())),
            io:nl(), % 改行のみを出力します。
            loop(Now) % 末尾再帰でループを実現します。
    end.

%%% 表示用プロセスを起動します。
start() ->
    register(display, spawn(fun() -> loop({0,0,0}) end)).

%%% 表示用プロセスの停止を指示します。
stop() ->
    display ! stop.

%%% 表示用プロセスにメッセージを送信します。
send(Cnt, Name, Addr, Port) ->
    display ! {Cnt, Name, Addr, Port}.
